import numpy as np
from numpy import ma as ma
from netCDF4 import Dataset as nc
import sys
from scipy.spatial import ckdtree


def getvarmask(varin):
    """define if the name of the mask corresponding to each variable"""
    if varin in ['avmu','un','ub','gru','gsu','gtu','ssu_m','u_io']:
        mask='umask' ; varlon='glamu' ; varlat='gphiu' 
    elif varin in ['avmv','vn','vb','grv','gsv','gtv','ssv_m','v_io']:
        mask='vmask' ; varlon='glamv' ; varlat='gphiv'  
    elif varin in ['rotb','rotn']:
        mask='fmask' ; varlon='glamf' ; varlat='gphif' 
    elif varin in ['avt','avmt','en','dissl','hdivb','hdivn','rhop','sn','tn','sb','tb','gcx','gcxb','gcxbfs','ssh_m','sshb','sshn','sss_m','sst_m','alb_ice','sst_io','sss_io']:
        mask='tmask' ; varlon='glamt' ; varlat='gphit' 
    else:
        print("variable not found")
        sys.exit(1)
    return(mask, varlon, varlat)
    
def read_files(fin, varin, fmin, fmout):
    """This function is reading the variables needed for the interpolation
    process from the input files
    Arguments:
    filein: Input restart type netcdf (already open)
    varin: Variable to interpolate
    varmask: name of the mask variable corresponding to varin 
            (tmask, umask, vmask, fmask)
    fmin: Input maskfile type netcdf (already open) 
    fmout: Output mask file type netcdf (already open)
    Outputs:
    fieldin: Field to interpolate
    depthin: 3D depth corresponding to fieldin
    thickin: 3D thichness corresponding to the fieldin
    thickout: 3D thickness toward which the interpolation 
              has to be performed
    depthout: 3D depth toward which the interpolation 
              has to be performed
    """
    
    #get the mask variable corresponding to varin
    varmask, varlon, varlat =getvarmask(varin)

    
    #read the input mask (needed because original s4 not masked)
    maskin=np.array(fmin.variables[varmask][:].squeeze())
    #generate a masked array to mask the land
    maskin=ma.array(maskin>=1,mask=maskin<0.5)
    #read the thickness
    thickin=np.array(fmin.variables["e3t_0"][:].squeeze())
    #read the depth
    depthin=np.array(fmin.variables["gdept_0"][:].squeeze())
    #read the longitude
    lonin=np.array(fmin.variables[varlon][:].squeeze())
    #read the latitude
    latin=np.array(fmin.variables[varlat][:].squeeze())

    
    fieldin=np.array(fin.variables[varin][:].squeeze())
    #diemnsion name
    dimnames=list(fin.variables[varin].dimensions)

    

    #if the last level of the input field is fully masked ()
    #or almost fully masked (less than 20points)
    #remove it
    #check if the variable is 2D or 3D:
    if len(fieldin.shape)==3:
        lastlev=maskin.shape[0]-1
        if (np.sum(maskin[lastlev,:,:])<=20) or (np.sum(maskin[lastlev,:,:]).mask):
            maskin=maskin[0:lastlev,:,:]
            fieldin=np.array(fieldin[0:lastlev,:,:])
            thickin=thickin[0:lastlev]
            depthin=depthin[0:lastlev]


    #read the mask field
    maskout=np.array(fmout.variables[varmask][:].squeeze())
    #thickness
    thickout=np.array(fmout.variables["e3t_0"][:].squeeze())
    #depth
    depthout=np.array(fmout.variables["gdept_0"][:].squeeze())
    #read the longitude
    lonout=np.array(fmout.variables[varlon][:].squeeze())
    #read the latitude
    latout=np.array(fmout.variables[varlat][:].squeeze())
    #generate a masked array to mask the land
    maskout=ma.array(maskout,mask=maskout<0.5)

    #permute the dimensions if needed
    # not take into account t dimension (size 1)
    if len(fieldin.shape)==3:
        correctdimorder=(u'z', u'y', u'x')
    else:
        correctdimorder=(u'y', u'x')

    dimnames.remove(u't')
    #the correctdimension order
    for idim,dimname in enumerate(dimnames):
        if dimname!=correctdimorder[idim]:
            print("Dimension"+dimname+"permuted")
            fieldin=moveaxis(fieldin, idim, 
                         correctdimorder.index(dimname))
    
    # extend the input depth in 3D 
    #(needed because from Nemo 3.6 both vary with time ans space)
    lz0, ly0, lx0 = maskin.shape
    lzout, lyout, lxout = maskout.shape
    if len(depthin.shape)==1:
        depthin.shape=(lz0,1,1)
        depthin=np.repeat(np.repeat(depthin,ly0, axis=1), lx0, axis=2)
    if len(depthout.shape)==1:
        depthout.shape=(lzout,1,1)
        depthout=np.repeat(np.repeat(depthout, ly0, axis=1), lx0, axis=2)
    # extend the input thickness in 3D
    if len(thickin.shape)==1:
        thickin.shape=(lz0,1,1)
        thickin=np.repeat(np.repeat(thickin,ly0, axis=1), lx0, axis=2)
    if len(thickout.shape)==1:
        thickout.shape=(lzout,1,1)
        thickout=np.repeat(np.repeat(thickout,ly0, axis=1), lx0, axis=2)


    #if the variable is 2d add one dimension of size one
    if len(fieldin.shape)==2:
        ly0,lx0=fieldin.shape
        maskin=maskin[0,:,:]
        thickin=np.zeros((ly0,lx0),'d')
        depthin=np.zeros((ly0,lx0),'d')
        thickout=np.zeros((ly0,lx0),'d')
        depthout=np.zeros((ly0,lx0),'d')
        maskout=maskout[0,:,:]
    
    #check that the mask corresponds to the input file
    #(in s4 restarts continents are filled with 0)
    if (fieldin[np.where(maskin.mask)]!=0).any():
        print("the variable: "+varmask)
        print("the input mask:")
        print(fmin.filepath())
        print("does not corespond to the input file:")
        print(fin.filepath())
        print("for the variable:"+varin)
        sys.exit(1)

    #apply the mask
    fieldin=ma.array(fieldin,mask=1-maskin)

    #check that the field is not empty
    if  (np.sum(fieldin)==0):
        print("The field written from the file: "+f)
        print("for the variable"+varin)
        print("mask with the file: "+fmask)
        print("seems to be empty")
        print("Please check your input")
        sys.exit(1)
    
    return(fieldin, thickin, depthin, lonin, latin, 
           maskout, thickout, depthout, lonout, latout)

def interp_vert(fieldin, thickin, depthin, thickout, depthout):
    """This function interpolates vertically a (x,y,z) field using the
    conservative method.
    Arguments:
    fieldin: Input Field type mask array
    maskout: output mask of the vertical interpolation
    History : Chloe Prodhomme - Initial version adpated
              from Cfutools extrap_vert function
              written by V. Guemas - 2016"""
    
    #name the dimensions' size
    (lz0,ly0,lx0)=fieldin.shape

    mask3D=1-fieldin.mask
    lzout=depthout.shape[0]

    #define the output files
    fieldout=np.zeros((lzout,ly0,lx0),'d')
    #

    #initialize auxiliary fields
    jkin=0
    maskout=np.zeros((lzout,ly0,lx0),'d')
    thick_jkout=np.zeros((ly0,lx0),'d')
    #loop over the output levels
    for jkout in np.arange(lzout) :
        #define the considered depth for the input and output field
        #depth of the beginning of the layer
        depthout0=depthout[jkout,:,:]-thickout[jkout,:,:]/2
        #depth of the end of the layer
        depthout1=depthout[jkout,:,:]+thickout[jkout,:,:]/2 
        #(for the last level take the last level again)
        depthin1=depthin[min(jkin+1,lz0-1),:,:]+thickin[min(jkin+1,lz0-1),:,:]/2  
        depthin0=depthin[min(jkin+1,lz0-1),:,:]-thickin[min(jkin+1,lz0-1),:,:]/2 
        #loop over the input level to find the correponding output depth
        while ( (jkin <= (lz0-1)) and (depthin1<=depthout1).any() ) :
            thick_bis=ma.array(np.minimum(thickin[jkin,:,:],
                                             depthin1-depthout0),
                               mask=fieldin.mask[jkin,:,:])
            fieldout[jkout,:,:]=fieldout[jkout,:,:]+np.multiply(fieldin[jkin,:,:],
                                                                thick_bis)
            thick_jkout=thick_jkout+thick_bis
            #look for the following level 
            jkin=jkin+1
            #actualize depthin with the new jkin
            #(for the last level take the last level again)
            jkinmin=min(jkin+1,lz0-1)
            depthin1=depthin[jkinmin,:,:]+thickin[jkinmin,:,:]/2  
            depthin0=depthin[jkinmin,:,:]-thickin[jkinmin,:,:]/2 

        jkinmin=min(jkin+1,lz0-1)
        thick_bis=ma.array(np.minimum(thickout[jkout,:,:],
                                         depthout1-depthin0), 
                            mask=fieldin.mask[jkinmin,:,:])
        
        fieldout[jkout,:,:]=fieldout[jkout,:,:]+np.multiply(fieldin[jkinmin
                                                                    ,:,:],
                                                            thick_bis)
        thick_jkout=thick_jkout+thick_bis            
        maskout[jkout,:,:]=thick_bis.mask
        fieldout[jkout,:,:]=fieldout[jkout,:,:]/thick_jkout
        thick_jkout=np.zeros((ly0,lx0),'d')
        
    #mask the output field
    fieldout=ma.array(fieldout, mask=maskout)
    return(fieldout) 

def lon_lat_to_cartesian(lon, lat, R=1):
    """calculates lon, lat coordinates of a point on a sphere with
    radius R
    """
    lon_r = np.radians(lon)
    lat_r = np.radians(lat)
    x = R * np.cos(lat_r) * np.cos(lon_r)
    y = R * np.cos(lat_r) * np.sin(lon_r)
    z = R * np.sin(lat_r)
    return x, y, z

def extrap(fieldin,lonin,latin,maskout):
    """This function extrapolates horizontally each level of a (x,y,z)
    field using the nearest neighbour method.
    Argument:
    fieldin: Input field (ma.array type)
    lonin: 2d longitude
    latin: 2d longitude
    maskout: mask of the field after extrapolation
    Output:
    fieldout: output field (ma.array)
    """
    if len(fieldin.shape)==2:
        ly0,lz0=fieldin.shape
        fieldin.shape=(1,ly0,lz0)
        maskout.shape=(1,ly0,lz0)

    #construct points to fill
    Pfill=ma.array((fieldin.mask)&(maskout!=0), mask=1-maskout)

    # fieldout intitialization
    fieldout=np.array(fieldin)

    #2d to 1d coordinates
    nlon = lonin.flatten()
    nlat = latin.flatten()
    
    #indexes of points to extrapolate
    indexes = np.where(Pfill) 
    #each level has a different mask, loop over the levels
    lz0=fieldin.shape[0]
        
    for lev in np.arange(lz0):
        newvar = ma.array(fieldin[lev,:,:], 
                          mask=fieldin.mask[lev,:,:])
        #not masked coordinates for the present level
        nomlon = lonin[np.where(newvar.mask==False)]
        nomlat = latin[np.where(newvar.mask==False)]
        #check if there is not-masked point at this level ## Chloe
        if (nomlon.shape)[0]!=0:
            #cartesian coordinates of not masked points
            levxs, levys, levzs = lon_lat_to_cartesian(nomlon, nomlat)
            levkdt = ckdtree.cKDTree(zip(levxs, levys, levzs))
            #points to extrapolate at the current level
            newindexes = np.where(indexes[0]==lev)
            #check if there is points to be filled ## Chloe
            if ((newindexes[0]).shape)[0]!=0:
                latidxs = indexes[1][newindexes]
                lonidxs = indexes[2][newindexes]
                exlon = nlon[lonidxs]
                exlat = nlat[latidxs]
                
                #cartesian coordinates of the points to extrapolate
                xt, yt, zt = lon_lat_to_cartesian(exlon, exlat)
                #calculates distances and indexes of the nearest neighbours
                dist, idx = levkdt.query(zip(xt, yt, zt), k=2)
                #assign to the result matrix the found values
                fieldout[lev,latidxs,
                         lonidxs] = newvar.flatten()[idx.T[1]]
    #mask field out
    fieldout=ma.array(fieldout,mask=1-maskout)
    return(fieldout)




def vertinterp_extrap(filein, varin, fmask, fmaskout):
    """read and interpolate verticaly and extrapolate horizontal the given variable
    from a restart file
    Arguments:
    filein: Input restart file
    varin: name of the variable to extrapolate
    fmask: mask file corresponding to varin
    fmaskout: mask to define the grid toward which the extrapolation if done
    """
    
    #read files and define variables needed for the rest of the process
    (fieldin, thickin, depthin, lonin, latin, 
     maskout, 
     thickout, depthout, 
     lonout, latout) = read_files(filein, 
                                varin, fmask, fmaskout)
    #check if the variable is 2D or 3D:
    if len(fieldin.shape)==3:
        #perform the vertical extrapolation
        fieldin=interp_vert(fieldin, thickin, depthin, thickout, depthout)
    
    #construct points to fill
    Pfill=ma.array((fieldin.mask)&(maskout!=0), mask=1-maskout)
    
    #find if the caspienne sea need to be filled
    caspbox=np.where((lonin>45)&(lonin<56)&(latin>35)&(latin<50))
    #longitude and latitude boundary of the caspienne sea
    loncasp1,loncasp2=caspbox[0].min(),caspbox[0].max()
    latcasp1,latcasp2=caspbox[1].min(),caspbox[1].max()
    #check there there no point to fill in the caspienne sea
    if len(fieldin.shape)==2:
        casptest=not(np.sum(Pfill[loncasp1:loncasp2,latcasp1:latcasp2], axis=(0,1)).mask)
    else: 
        casptest=not(np.sum(Pfill[:,loncasp1:loncasp2,latcasp1:latcasp2], axis=(0,1,2)).mask)
    if  casptest:
        print("you need a climatology to fill the file")
        print("will be handle in the coming version")
        sys.exit(1)

    
    #horizontal extrapolation
    fieldout=extrap(fieldin,lonin,latin,maskout)
    return(fieldout)



def vertextp_extrap_fill(filein, maskin, maskout):
    """This script interpolates vertically, extrapolates horizontally,
    fills up empty seas with climatologies, extrapolate vertically  ocean restarts 
    of s4 to be able to run EC-Earth."""
    
    listvar3d=['en','ub','vb','tb','sb','rotb','hdivb','un','vn','tn','sn',
                   'rotn','hdivn', 'gcx','gcxb','sshb','sshn']

    
    
